@@ -1,9 +1,7 @@
-from __future__ import print_function
-
 import ast
 from ast import parse
 import codecs
-import collections
+import dataclasses as dc
 import difflib
 import importlib
 import io
@@ -11,6 +9,13 @@ import os
 import re
 import sys
 import time
+from typing import Any
+from typing import Iterable
+from typing import List
+from typing import NamedTuple
+from typing import Optional
+from typing import Set
+from typing import Tuple
 
 import flake8_import_order as f8io
 from flake8_import_order.styles import lookup_entry_point
@@ -21,106 +26,185 @@ from .vendored.flake8 import matches_filename
 from .vendored.flake8 import normalize_path
 
 
-def _rewrite_source(options, filename, source_lines):
+@dc.dataclass
+class Rewriter:
+    options: Any
+    filename: str
+    source_lines: List[str]
 
-    keep_threshhold = options.heuristic_unused
-    expand_stars = options.expand_stars
-    style_entry_point = lookup_entry_point(options.style)
-    style = style_entry_point.load()
+    def __post_init__(self):
+        self.keep_threshhold: float = self.options.heuristic_unused
+        self.expand_stars: bool = self.options.expand_stars
+        style_entry_point = lookup_entry_point(self.options.style)
+        self.style = style_entry_point.load()
 
-    stats = {
-        "starttime": time.time(),
-        "names_from_star": 0,
-        "star_imports_removed": 0,
-        "removed_imports": 0,
-    }
+        self.stats = {
+            "starttime": time.time(),
+            "names_from_star": 0,
+            "star_imports_removed": 0,
+            "removed_imports": 0,
+        }
 
-    # parse the code.  get the imports and a collection of line numbers
-    # we definitely don't want to discard
-    imports, _, lines_with_code = _parse_toplevel_imports(
-        options, filename, source_lines
-    )
-
-    original_imports = len(imports)
-    if imports:
-        imports_start_on = imports[0].lineno
-    else:
-        imports_start_on = 0
-
-    # assemble a set of line numbers that will not be copied to the
-    # output.  E.g. lines where import statements occurred, or the
-    # extra lines they take up which we figure out by looking at the
-    # "gap" between statements
-    import_gap_lines = _get_import_discard_lines(
-        filename, source_lines, imports, lines_with_code
-    )
-
-    # flatten imports into single import per line and rewrite
-    # full source
-    if not options.multi_imports:
-        imports = list(
-            _dedupe_single_imports(
-                _as_single_imports(imports, stats, expand_stars=expand_stars),
-                stats,
-            )
+    def _do_rewrite(self, source_lines: List[str], type_check_pass: bool):
+        if type_check_pass:
+            # Stats are collected only on the non type check pass.
+            stats = self.stats.copy()
+            type_checking_blocks = TypeCheckingBlocks(source_lines)
+        else:
+            stats = self.stats
+            type_checking_blocks = None
+        # parse the code.  get the imports and a collection of line numbers
+        # we definitely don't want to discard
+        imports, _, lines_with_code = _parse_toplevel_imports(
+            self.options, self.filename, source_lines, type_checking_blocks
         )
 
-    on_source_lines = _write_source(
-        source_lines, imports, [], import_gap_lines, imports_start_on, style
-    )
-    # now parse again.  Because pyflakes won't tell us about unused
-    # imports that are not the first import, we had to flatten first.
-    imports, warnings, lines_with_code = _parse_toplevel_imports(
-        options, filename, on_source_lines, drill_for_warnings=True
-    )
+        original_imports = len(imports)
+        if imports:
+            imports_start_on = imports[0].lineno
+        else:
+            imports_start_on = 0
 
-    # now remove unused names from the imports
-    # if number of imports is greater than keep_threshold% of the total
-    # lines of code, don't remove names, assume this is like a
-    # package file
-    if not lines_with_code:
-        stats["import_proportion"] = import_proportion = 0
-    else:
-        stats["import_proportion"] = import_proportion = (
-            (
-                len(imports)
-                + stats["star_imports_removed"]
-                - stats["names_from_star"]
+        # assemble a set of line numbers that will not be copied to the
+        # output.  E.g. lines where import statements occurred, or the
+        # extra lines they take up which we figure out by looking at the
+        # "gap" between statements
+        import_gap_lines: Set[int] = _get_import_discard_lines(
+            source_lines, imports, lines_with_code
+        )
+
+        # flatten imports into single import per line and rewrite
+        # full source
+        if not self.options.multi_imports:
+            imports = list(
+                _dedupe_single_imports(
+                    _as_single_imports(
+                        imports, stats, expand_stars=self.expand_stars
+                    ),
+                    stats,
+                )
             )
-            / float(len(lines_with_code))
-        ) * 100
 
-    if keep_threshhold is None or import_proportion < keep_threshhold:
-        _remove_unused_names(imports, warnings, stats)
+        on_source_lines = _write_source(
+            source_lines,
+            imports,
+            [],
+            import_gap_lines,
+            imports_start_on,
+            self.style,
+        )
+        if type_check_pass:
+            TypeCheckingBlocks(on_source_lines).remove_empty_blocks(
+                on_source_lines
+            )
+            type_checking_blocks = TypeCheckingBlocks(on_source_lines)
+        # now parse again.  Because pyflakes won't tell us about unused
+        # imports that are not the first import, we had to flatten first.
+        imports, warnings, lines_with_code = _parse_toplevel_imports(
+            self.options,
+            self.filename,
+            on_source_lines,
+            type_checking_blocks,
+            drill_for_warnings=True,
+        )
 
-    stats["import_line_delta"] = len(imports) - original_imports
+        if type_check_pass:
+            # now remove unused names from the imports if keep unused was not
+            # specified in the arguments
+            if self.options.keep_unused_type_checking is False:
+                _remove_unused_names(imports, warnings, stats)
+        else:
+            # now remove unused names from the imports
+            # if number of imports is greater than keep_threshold% of the total
+            # lines of code, don't remove names, assume this is like a
+            # package file
+            if not lines_with_code:
+                stats["import_proportion"] = import_proportion = 0
+            else:
+                stats["import_proportion"] = import_proportion = (
+                    (
+                        len(imports)
+                        + stats["star_imports_removed"]
+                        - stats["names_from_star"]
+                    )
+                    / float(len(lines_with_code))
+                ) * 100
 
-    sorted_imports, nosort_imports = sort_imports(style, imports, options)
+            if (
+                self.keep_threshhold is None
+                or import_proportion < self.keep_threshhold
+            ):
+                _remove_unused_names(imports, warnings, stats)
 
-    rewritten = _write_source(
-        source_lines,
-        sorted_imports,
-        nosort_imports,
-        import_gap_lines,
-        imports_start_on,
-        style,
-    )
+        stats["import_line_delta"] = len(imports) - original_imports
 
-    differ = list(difflib.Differ().compare(source_lines, rewritten))
+        sorted_imports, nosort_imports = sort_imports(
+            self.style, imports, self.options
+        )
 
-    stats["added"] = len([l for l in differ if l.startswith("+ ")])
-    stats["removed"] = len([l for l in differ if l.startswith("- ")])
-    stats["is_changed"] = bool(stats["added"] or stats["removed"])
-    stats["totaltime"] = time.time() - stats["starttime"]
-    return rewritten, stats
+        rewritten = _write_source(
+            source_lines,
+            sorted_imports,
+            nosort_imports,
+            import_gap_lines,
+            imports_start_on,
+            self.style,
+        )
+        if type_check_pass:
+            TypeCheckingBlocks(rewritten).remove_empty_blocks(rewritten)
+        return rewritten
+
+    def rewrite(self):
+        # does two passes: first one handles only the type checking imports;
+        # second one handles the top level imports instead.
+        rewritten = self._do_rewrite(self.source_lines, type_check_pass=True)
+        rewritten = self._do_rewrite(rewritten, type_check_pass=False)
+
+        differ = list(difflib.Differ().compare(self.source_lines, rewritten))
+
+        self.stats["added"] = len([l for l in differ if l.startswith("+ ")])
+        self.stats["removed"] = len([l for l in differ if l.startswith("- ")])
+        self.stats["is_changed"] = bool(
+            self.stats["added"] or self.stats["removed"]
+        )
+        self.stats["totaltime"] = time.time() - self.stats["starttime"]
+        return rewritten, self.stats
+
+
+class TypeCheckingBlocks:
+    def __init__(self, source_lines):
+        self.type_checking_blocks = []
+        in_type_checking_block = False
+        for lineno, line in enumerate(source_lines, 1):
+            if re.match(r"^if [\w+\.]*TYPE_CHECKING:", line):
+                # we are inside an "if TYPE_CHECKING:" block
+                in_type_checking_block = True
+                self.type_checking_blocks.append((lineno, set()))
+            elif in_type_checking_block:
+                if line and not re.match(r"^\s+(from|import)", line):
+                    # line is not empty and not an indented import so the
+                    # type checking import block has ended
+                    in_type_checking_block = False
+                else:
+                    self.type_checking_blocks[-1][1].add(lineno)
+
+    def __contains__(self, lineno):
+        return any(lineno in lines for _, lines in self.type_checking_blocks)
+
+    def remove_empty_blocks(self, source_lines):
+        for block, lines in self.type_checking_blocks:
+            if all(not source_lines[line - 1] for line in lines):
+                source_lines[block - 1] = ""
 
 
 def _get_import_discard_lines(
-    filename, source_lines, imports, lines_with_code
+    source_lines: List[str],
+    imports: List["ClassifiedImport"],
+    lines_with_code: Set[str],
 ):
     """Get line numbers that are part of imports but not in the AST."""
 
-    import_gap_lines = {node.lineno for node in imports}
+    import_gap_lines: Set[int] = {node.lineno for node in imports}
 
     intermediary_whitespace_lines = []
 
@@ -164,14 +248,14 @@ def _is_whitespace_or_comment(line):
 
 
 def _write_source(
-    source_lines,
-    imports,
+    source_lines: List[str],
+    imports: List["ClassifiedImport"],
     nosort_imports,
-    import_gap_lines,
-    imports_start_on,
-    style,
+    import_gap_lines: Set[int],
+    imports_start_on: int,
+    style: Any,
 ):
-    buf = []
+    buf: List[str] = []
     previous_import = None
     for lineno, line in enumerate(source_lines, 1):
         if lineno == imports_start_on:
@@ -205,13 +289,15 @@ def _write_import(import_node):
     modules.sort(key=lambda x: x.lower())
     modules = ", ".join(modules)
     if not import_node.is_from:
-        return "import %s%s%s" % (
+        return "%simport %s%s%s" % (
+            " " * import_node.col_offset,
             modules,
             "  # noqa" if import_node.noqa else "",
             " nosort" if import_node.nosort else "",
         )
     else:
-        return "from %s%s import %s%s%s" % (
+        return "%sfrom %s%s import %s%s%s" % (
+            " " * import_node.col_offset,
             "." * import_node.level,
             import_node.modules[0] or "",
             modules,
@@ -220,24 +306,20 @@ def _write_import(import_node):
         )
 
 
-class ClassifiedImport(
-    collections.namedtuple(
-        "ClassifiedImport",
-        [
-            "type",
-            "is_from",
-            "modules",
-            "names",
-            "lineno",
-            "level",
-            "package",
-            "ast_names",
-            "render_ast_names",
-            "noqa",
-            "nosort",
-        ],
-    )
-):
+class ClassifiedImport(NamedTuple):
+    type: f8io.ImportType
+    is_from: bool
+    modules: list
+    names: list
+    lineno: int
+    col_offset: int
+    level: int
+    package: str
+    ast_names: Optional[str]
+    render_ast_names: list
+    noqa: bool
+    nosort: bool
+
     def __hash__(self):
         return hash((self.type, self.is_from, self.lineno))
 
@@ -274,12 +356,18 @@ class ClassifiedImport(
 
 class ImportVisitor(f8io.ImportVisitor):
     def __init__(
-        self, source_lines, application_import_names, application_package_names
+        self,
+        source_lines,
+        application_import_names,
+        application_package_names,
+        type_checking_blocks,
     ):
-        self.imports = []
+        self.imports: List[ClassifiedImport] = []
         self.source_lines = source_lines
         self.application_import_names = frozenset(application_import_names)
         self.application_package_names = frozenset(application_package_names)
+        self.type_checking_blocks = type_checking_blocks
+        self.top_level = type_checking_blocks is None
 
     def _get_flags(self, lineno):
         line = self.source_lines[lineno - 1].rstrip()
@@ -291,8 +379,13 @@ class ImportVisitor(f8io.ImportVisitor):
                 nosort = True
         return noqa, nosort
 
+    def _check_node(self, node):
+        return (self.top_level and node.col_offset == 0) or (
+            not self.top_level and node.lineno in self.type_checking_blocks
+        )
+
     def visit_Import(self, node):  # noqa: N802
-        if node.col_offset == 0:
+        if self._check_node(node):
             modules = [alias.name for alias in node.names]
             types_ = {self._classify_type(module) for module in modules}
             if len(types_) == 1:
@@ -306,6 +399,7 @@ class ImportVisitor(f8io.ImportVisitor):
                 modules,
                 [],
                 node.lineno,
+                node.col_offset,
                 0,
                 f8io.root_package_name(modules[0]),
                 node.names,
@@ -316,7 +410,7 @@ class ImportVisitor(f8io.ImportVisitor):
             self.imports.append(classified_import)
 
     def visit_ImportFrom(self, node):  # noqa: N802
-        if node.col_offset == 0:
+        if self._check_node(node):
             module = node.module or ""
             if node.level > 0:
                 type_ = f8io.ImportType.APPLICATION_RELATIVE
@@ -330,6 +424,7 @@ class ImportVisitor(f8io.ImportVisitor):
                 [module],
                 names,
                 node.lineno,
+                node.col_offset,
                 node.level,
                 f8io.root_package_name(module),
                 node.names,
@@ -341,7 +436,11 @@ class ImportVisitor(f8io.ImportVisitor):
 
 
 def _parse_toplevel_imports(
-    options, filename, source_lines, drill_for_warnings=False
+    options: Any,
+    filename: str,
+    source_lines: List[str],
+    type_checking_blocks: Optional[TypeCheckingBlocks],
+    drill_for_warnings: bool = False,
 ):
     source = "\n".join(source_lines)
 
@@ -355,7 +454,7 @@ def _parse_toplevel_imports(
 
     if drill_for_warnings:
         warnings_set = _drill_for_warnings(
-            options, filename, source_lines, warnings
+            options, filename, source_lines, warnings, type_checking_blocks
         )
     else:
         warnings_set = None
@@ -364,13 +463,20 @@ def _parse_toplevel_imports(
         source_lines,
         options.application_import_names.split(","),
         options.application_package_names.split(","),
+        type_checking_blocks,
     )
     f8io_visitor.visit(tree)
     imports = f8io_visitor.imports
     return imports, warnings_set, lines_with_code
 
 
-def _drill_for_warnings(options, filename, source_lines, warnings):
+def _drill_for_warnings(
+    options: Any,
+    filename: str,
+    source_lines: List[str],
+    warnings: pyflakes.checker.Checker,
+    type_checking_blocks: Optional[TypeCheckingBlocks],
+):
     # pyflakes doesn't warn for all occurrences of an unused import
     # if that same symbol is repeated, so run over and over again
     # until we find every possible warning.  assumes single-line
@@ -384,8 +490,9 @@ def _drill_for_warnings(options, filename, source_lines, warnings):
                 ignore_errors.update(codes)
 
     source_lines = list(source_lines)
-    warnings_set = set()
+    warnings_set: Set[Tuple[str, int]] = set()
     seen_lineno = set()
+    top_level = type_checking_blocks is None
     while True:
         has_warnings = False
         for warning in warnings.messages:
@@ -398,13 +505,15 @@ def _drill_for_warnings(options, filename, source_lines, warnings):
             if "F401" in ignore_errors:
                 continue
 
-            # we only deal with "top level" imports for now. imports
-            # inside of conditionals or in defs aren't counted.
-            whitespace = re.match(
-                r"^\s*", source_lines[warning.lineno - 1]
-            ).group(0)
-            if whitespace:
-                continue
+            line = source_lines[warning.lineno - 1]
+            if top_level:
+                # when dealing with "top level" imports, imports
+                # inside of conditionals or in defs aren't counted.
+                if re.match(r"^\s*", line).group(0):
+                    continue
+            else:
+                if warning.lineno not in type_checking_blocks:
+                    continue
             has_warnings = True
             warnings_set.add((warning.message_args[0], warning.lineno))
 
@@ -417,6 +526,8 @@ def _drill_for_warnings(options, filename, source_lines, warnings):
         if not has_warnings:
             break
 
+        if type_checking_blocks:
+            type_checking_blocks.remove_empty_blocks(source_lines)
         source = "\n".join(source_lines)
         tree = ast.parse(source, filename)
         warnings = pyflakes.checker.Checker(tree, filename)
@@ -424,7 +535,11 @@ def _drill_for_warnings(options, filename, source_lines, warnings):
     return warnings_set
 
 
-def _remove_unused_names(imports, warnings, stats):
+def _remove_unused_names(
+    imports: List[ClassifiedImport],
+    warnings: Set[Tuple[str, int]],
+    stats: dict,
+):
     noqa_lines = set(
         import_node.lineno for import_node in imports if import_node.noqa
     )
@@ -455,10 +570,12 @@ def _remove_unused_names(imports, warnings, stats):
     imports[:] = new_imports
 
 
-def _dedupe_single_imports(import_nodes, stats):
+def _dedupe_single_imports(
+    import_nodes: Iterable[ClassifiedImport], stats: dict
+):
 
     seen = {}
-    orig_order = []
+    orig_order: List[Tuple[ClassifiedImport, Any]] = []
 
     for import_node in import_nodes:
         if not import_node.is_from:
@@ -491,7 +608,11 @@ def _dedupe_single_imports(import_nodes, stats):
             stats["removed_imports"] += 1
 
 
-def _as_single_imports(import_nodes, stats, expand_stars=False):
+def _as_single_imports(
+    import_nodes: List[ClassifiedImport],
+    stats: dict,
+    expand_stars: bool = False,
+):
 
     for import_node in import_nodes:
         if not import_node.is_from:
@@ -502,6 +623,7 @@ def _as_single_imports(import_nodes, stats, expand_stars=False):
                     [ast_name.name],
                     [],
                     import_node.lineno,
+                    import_node.col_offset,
                     import_node.level,
                     import_node.package,
                     [ast_name],
@@ -523,6 +645,7 @@ def _as_single_imports(import_nodes, stats, expand_stars=False):
                             import_node.modules,
                             [star_name],
                             import_node.lineno,
+                            import_node.col_offset,
                             import_node.level,
                             import_node.package,
                             [ast_cls(star_name, asname=None)],
@@ -537,6 +660,7 @@ def _as_single_imports(import_nodes, stats, expand_stars=False):
                         import_node.modules,
                         [ast_name.name],
                         import_node.lineno,
+                        import_node.col_offset,
                         import_node.level,
                         import_node.package,
                         [ast_name],
@@ -546,7 +670,7 @@ def _as_single_imports(import_nodes, stats, expand_stars=False):
                     )
 
 
-def sort_imports(style, imports, options):
+def sort_imports(style: Any, imports: List[ClassifiedImport], options: Any):
     tosort = []
     nosort = []
 
@@ -650,7 +774,8 @@ def _run_file(options, filename):
                 "keep-unused and heuristic-unused are mutually exclusive"
             )
         options.heuristic_unused = 0
-    result, stats = _rewrite_source(options, filename, source_lines)
+        options.keep_unused_type_checking = True
+    result, stats = Rewriter(options, filename, source_lines).rewrite()
     totaltime = stats["totaltime"]
     if not stats["is_changed"]:
         sys.stderr.write(
